"""
A collection of simple methods for consistently creating identifiers for things

Author: Ilya Baldin (ibaldin@renci.org), Xi Yang (xiyang@es.net)
"""
from typing import Tuple


def mac_to_node_id(mac: str) -> str:
    """
    turn mac address into a unique string to act
    as node id, but don't make it look like mac address
    :param mac:
    :return:
    """

    return 'node_id-' + mac.replace(':', '-')


def dp_switch_name_id(site: str, ip: str) -> Tuple[str, str]:
    """
    Return a tuple of name and id of a switch name and node id
    :param site:
    :param ip:
    :return:
    """
    name = site.lower() + '-data-sw'
    return name,  'node+' + name + ':ip+' + ip


def dp_port_id(switch: str, port: str) -> str:
    """
    Return a unique id of a DP switch port based on switch name and port name
    :param switch:
    :param port:
    :return:
    """
    return 'port+' + switch + ':' + port


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author Ilya Baldin (ibaldin@renci.org)
from typing import Dict, Any

import json

from .network_node import NodeSliver
from .network_service import NetworkServiceSliver
from fim.graph.abc_property_graph import ABCPropertyGraph


class JSONSliver:
    """
    This class uses ABC Property Graph dictionary serialization
    to also convert slivers to and from JSON. It is made into a
    separate class to avoid circular dependencies.
    """

    @staticmethod
    def sliver_to_json(sliver) -> str:
        """
        Take an object that is a descendant of a BaseSliver and
        convert it to JSON
        :param sliver:
        :return:
        """
        d = ABCPropertyGraph.sliver_to_dict(sliver)
        if d is not None:
            return json.dumps(d)

    @staticmethod
    def node_sliver_from_json(s: str) -> NodeSliver:
        """
        Recover a node sliver from JSON
        :param s:
        :return:
        """
        d = json.loads(s)
        if not isinstance(d, dict):
            raise RuntimeError(f'Unable to decode NodeSliver from {s=}')
        return ABCPropertyGraph.build_deep_node_sliver_from_dict(props=d)

    @staticmethod
    def network_service_sliver_from_json(s: str) -> NetworkServiceSliver:
        """
        Recover a network service sliver from JSON
        :param s:
        :return:
        """
        d = json.loads(s)
        if not isinstance(d, dict):
            raise RuntimeError(f'Unable to decode NetworkServiceSliver from {s=}')
        return ABCPropertyGraph.build_deep_ns_sliver_from_dict(props=d)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

"""
This utility implements a number of operations to make it easier to develop and debug
information models in FABRIC. It

- takes in a serialized model in GraphML (typically produced by yEd) and
adds various identifiers (node, graph, etc) as necessary for models to become readable
by CF actors.
- loads a model from a file into an indicated Neo4j instance
- remove a specific graph or all graphs from the database

Configuration is done using a yaml file named by default fim_config.yaml structured as:

# default fim_config configuration file for FIM utilities
neo4j:
  url: neo4j://0.0.0.0:7687
  user: neo4j
  pass: password
  import_host_dir: /host/directory/seen/by/neo4j/docker/as/imports
  import_dir: /imports

Designed to work with Neo4j-APOC Docker
https://hub.docker.com/repository/docker/fabrictestbed/neo4j-apoc
"""

import argparse
import sys
import yaml
import logging
import traceback

from fim.graph.abc_property_graph import ABCPropertyGraph
from fim.graph.neo4j_property_graph import Neo4jPropertyGraph, Neo4jGraphImporter, PropertyGraphImportException
from fim.graph.resources.neo4j_arm import Neo4jARMGraph
from fim.graph.resources.neo4j_cbm import Neo4jCBMGraph
from fim.pluggable import PluggableRegistry

FIM_CONFIG_YAML = "fim_config.yaml"


def load_file(*, filename, graph_id, neo4j_config):
    """
    load graph from file
    :param filename:
    :param graph_id:
    :param neo4j_config:
    :return:
    """
    neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                              pswd=neo4j_config["pass"],
                                              import_host_dir=neo4j_config["import_host_dir"],
                                              import_dir=neo4j_config["import_dir"])
    logging.info(f"Loading graph into Neo4j")
    _graph = neo4j_graph_importer.import_graph_from_file(graph_file=filename, graph_id=graph_id)
    logging.info(f"Created graph {_graph}")
    try:
        _graph.validate_graph()
        return _graph.graph_id
    except PropertyGraphImportException as pe:
        logging.info(f"Unable to load graph due to error {pe.msg}, deleting")
        _graph.delete_graph()
        return None


def load_file_direct(*, filename, neo4j_config):
    """
    Load specified file directly with no manipulations or validation
    :param filename:
    :param neo4j_config:
    :return:
    """
    neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                              pswd=neo4j_config["pass"],
                                              import_host_dir=neo4j_config["import_host_dir"],
                                              import_dir=neo4j_config["import_dir"])
    return neo4j_graph_importer.import_graph_from_file_direct(graph_file=filename).graph_id


def enumerate_nodes(*, filename, new_filename, neo4j_config):
    """
    Add unique NodeId property to each node in the graph and save under
    a new name
    :param filename:
    :param new_filename:
    :param neo4j_config
    :return:
    """
    neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                              pswd=neo4j_config["pass"],
                                              import_host_dir=neo4j_config["import_host_dir"],
                                              import_dir=neo4j_config["import_dir"])
    neo4j_graph_importer.enumerate_graph_nodes(graph_file=filename, new_graph_file=new_filename)


def delete_graphs(*, neo4j_config, graph_id=None):
    """
    Delete a specific graph or all graphs if graph_id is none
    :param neo4j_config
    :param graph_id:
    :return:
    """
    neo4j_graph = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                     pswd=neo4j_config["pass"],
                                     import_host_dir=neo4j_config["import_host_dir"],
                                     import_dir=neo4j_config["import_dir"])
    if graph_id is None:
        neo4j_graph.delete_all_graphs()
    else:
        neo4j_graph.delete_graph(graph_id=graph_id)


def merge_ads(*, neo4j_config, file_names, delete_arms, delete_adms):
    """
    Merge these advertisements together in the order provided
    :param neo4j_config:
    :param file_names: list of file names
    :return:
    """
    neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                              pswd=neo4j_config["pass"],
                                              import_host_dir=neo4j_config["import_host_dir"],
                                              import_dir=neo4j_config["import_dir"])
    cbm = Neo4jCBMGraph(importer=neo4j_graph_importer)
    logging.info(f'Created blank CBM {cbm.graph_id}')

    arms = dict()
    for file_name in file_names:
        _graph = neo4j_graph_importer.import_graph_from_file(graph_file=file_name)
        logging.info(f'Loaded {file_name} as graph ID {_graph.graph_id}')
        _graph.validate_graph()
        if arms.get(file_name, None) is not None:
            logging.error(f'Encountered model {file_name} more than once, please revise parameters.')
            sys.exit(-1)
        arms[file_name] = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=_graph.graph_id,
                                                                 importer=neo4j_graph_importer))
        # generate a dict of ADMs from site graph ARM
        site_adms = arms[file_name].generate_adms()
        logging.info(f'ADMS for {file_name}:')
        for adm_id in site_adms.keys():
            logging.info(f'  ADM id {adm_id}: {site_adms[adm_id].graph_id}')

        # desired ADM is under 'primary'
        site_adm = site_adms['primary']

        logging.info(f'Merging into CBM')
        cbm.merge_adm(adm=site_adm)
        cbm.validate_graph()

        if delete_arms:
            logging.info(f'Deleting ARM {file_name}')
            arms[file_name].delete_graph()
        if delete_adms:
            logging.info(f'Deleting ADMs for {file_name}')
            for adm in site_adms.values():
                adm.delete_graph()
    logging.info('Completed')


def save_graph(*, outfile, graph_id, neo4j_config):
    """
    Save indicated graph id into a file
    :param outfile:
    :param graph_id:
    :param new4j_config:
    :return:
    """
    neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                              pswd=neo4j_config["pass"],
                                              import_host_dir=neo4j_config["import_host_dir"],
                                              import_dir=neo4j_config["import_dir"])
    graph = Neo4jPropertyGraph(graph_id=graph_id, importer=neo4j_graph_importer)
    graph_string = graph.serialize_graph()
    with open(outfile, 'w') as f:
        f.write(graph_string)


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    # split into different mutually exclusive operations

    operations = parser.add_mutually_exclusive_group()
    operations.add_argument("-l", "--load", action="store_true",
                            help="load specified file into Neo4j (requires one or more -f, optional -g, -r)")
    parser.add_argument("-m", "--merge", action="store_true",
                        help="generate ADMs from provided ARMs and merge into CBM")
    operations.add_argument("-w", "--wipe", action="store_true",
                            help="wipe neo4j instance of all graphs (optional -g)")
    operations.add_argument("-e", "--enumerate", action="store_true",
                            help="enumerate all nodes in a graph in specified file"
                                 " assigning them unique GUIDs and save (requires -f and -o)")
    operations.add_argument("-s", "--save", action="store_true",
                            help="save a specific graph into a file (requires -o and -g)")
    parser.add_argument("-c", "--config", action="store", default=FIM_CONFIG_YAML,
                        help="provide alternative configuration file")
    parser.add_argument("-f", "--file", action="append",
                        help="GraphML file(s) to be operated on")
    parser.add_argument("-o", "--outfile", action="append",
                        help="output GraphML file(s)")
    parser.add_argument("-g", "--graph", action="append",
                        help="identifier of the graph in the Neo4j database")
    parser.add_argument("-r", "--direct", action="store_true",
                        help="load file directly without validation or any manipulations")
    parser.add_argument("-d", "--debug", action="count",
                        help="turn on debugging")

    args = parser.parse_args()

    if args.debug is None:
        logging.basicConfig(level=logging.INFO)
    elif args.debug >= 1:
        logging.basicConfig(level=logging.DEBUG)

    try:
        with open(args.config, 'r') as config_file:
            yaml_config = yaml.safe_load(config_file)
    except IOError:
        logging.error(f"Unable to open config file {args.config}, exiting", file=sys.stderr)
        sys.exit(-1)

    if args.load:
        if args.file is None or len(args.file) == 0:
            logging.error("Must specify at least one -f option", file=sys.stderr)
            sys.exit(-1)
        if not args.direct and (args.graph is None or len(args.graph) != len(args.file)):
            logging.error("Must specify the same number of -g options as -f")
            sys.exit(-1)
        # load a file with a given id (or generated id)
        logging.info(f'Loading {args.file}')
        try:
            if not args.direct:
                for file, graph in zip(args.file, args.graph):
                    gid = load_file(filename=file, neo4j_config=yaml_config["neo4j"],
                                    graph_id=args.graph)
                    if gid is not None:
                        logging.info(f"Graph {file} successfully loaded with id {gid}")
                    else:
                        sys.exit(-1)
            else:
                for file in args.file:
                    gid = load_file_direct(filename=file, neo4j_config=yaml_config["neo4j"])
                    if gid is not None:
                        logging.info(f"Graph {file} successfully loaded with id {gid}")
                    else:
                        sys.exit(-1)
        except Exception as e:
            logging.error(f"Unable to load file {args.file} to Neo4j due to {e.args}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            sys.exit(-1)
    elif args.merge:
        if args.file is None or len(args.file) == 0:
            logging.error('Must specify at least one -f option', file=sys.stderr)
            sys.exit(-1)
        # merge specified files into a CBM
        try:
            merge_ads(neo4j_config=yaml_config["neo4j"], file_names=args.file,
                      delete_arms=True, delete_adms=True)
        except Exception as e:
            logging.error(f"Unable to merge files {args.file} to Neo4j due to {e.args}", file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            sys.exit(-1)
    elif args.wipe:
        # wipe one graph or whole database
        delete_graphs(neo4j_config=yaml_config["neo4j"], graph_id=args.graph)
        if args.graph is not None:
            logging.info(f"Deleted graph {args.graph} from the database")
        else:
            logging.info("Deleted all graphs from the database")
    elif args.enumerate:
        if args.file is None or len(args.file) == 0:
            logging.error('Must specify at least one -f option', file=sys.stderr)
            sys.exit(-1)
        if args.outfile is None or len(args.outfile) != len(args.file):
            logging.error('Must specify same number of -o options as -f options', file=sys.stderr)
            sys.exit(-1)
        # enumerate a file
        for infile, outfile in zip(args.file, args.outfile):
            logging.info(f"Enumerating nodes in {infile} and saving as {outfile}")
            enumerate_nodes(filename=infile, new_filename=outfile, neo4j_config=yaml_config["neo4j"])
    elif args.save:
        if args.outfile is None or len(args.outfile) != 1:
            logging.error('Must specify one -o option', file=sys.stderr)
            sys.exit(-1)
        # save graph
        outfile = args.outfile[0]
        graph = args.graph
        if graph is None:
            logging.error("Must specify graph id (-g)", file=sys.stderr)
            sys.exit(-1)

        logging.info(f"Saving graph {graph} into file {outfile}")
        save_graph(outfile=outfile, graph_id=graph, neo4j_config=yaml_config["neo4j"])
    else:
        logging.error("Please specify one of -h, -l, -e or -w", file=sys.stderr)



import unittest

import fim.user as fu


def load(file_name: str):

    t = fu.ExperimentTopology()
    t.load(file_name=file_name)
    return t


class LoadTests(unittest.TestCase):

    def testNetworkXLoad(self):
        """
        Test for a stupid case where loading silently failed on duplicate graph IDs
        """
        print('')
        t1 = load(file_name="test/before.graphml")
        print(f"{t1.nodes['n2'].management_ip=}")
        t2 = load(file_name="test/after.graphml")
        print(f"{t2.nodes['n2'].management_ip=}")
        self.assertEqual(str(t2.nodes['n2'].management_ip), '128.163.179.51', "No management IP address found")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

"""
This collection of classes helps collect authorization attributes about various resources
"""

from typing import Dict, Any, List
from collections import defaultdict
from datetime import datetime, timedelta, timezone
import json

from fim.user.topology import ExperimentTopology
from fim.user.node import Node
from fim.user.network_service import NetworkService, ServiceType
from fim.graph.slices.networkx_asm import NetworkxASM
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver
from fim.view_only_dict import ViewOnlyDict


class ResourceAuthZAttributes:

    # standard resource attributes we are able to send to PDP
    RESOURCE_TYPE = "urn:fabric:xacml:attributes:resource-type"
    RESOURCE_CPU = "urn:fabric:xacml:attributes:resource-cpu"
    RESOURCE_RAM = "urn:fabric:xacml:attributes:resource-ram"
    RESOURCE_DISK = "urn:fabric:xacml:attributes:resource-disk"
    RESOURCE_BW = "urn:fabric:xacml:attribute:resource-bw"
    RESOURCE_SITE = "urn:fabric:xacml:attribute:resource-site"
    RESOURCE_COMPONENT = "urn:fabric:xacml:attribute:resource-component"
    RESOURCE_PEER_SITE = "urn:fabric:xacml:attribute:resource-peersite"
    RESOURCE_MIRROR_SITE = "urn:fabric:xacml:attribute:resource-mirrorsite"
    RESOURCE_FACILITY_PORT = "urn:fabric:xacml:attribute:resource-facility-port"
    RESOURCE_MEASUREMENTS = "urn:fabric:xacml:attribute:resource-with-measurements"
    RESOURCE_LIFETIME = "urn:fabric:xacml:attributes:resource-lifetime"
    RESOURCE_PROJECT = "urn:fabric:xacml:attributes:resource-project"
    RESOURCE_SUBJECT = "urn:fabric:xacml:attributes:resource-subject"
    ACTION_ID = "urn:oasis:names:tc:xacml:1.0:action:action-id"
    SUBJECT_ID = "urn:oasis:names:tc:xacml:1.0:subject:subject-id"
    SUBJECT_PROJECT = "urn:fabric:xacml:attributes:subject-project"
    PROJECT_TAG = "urn:fabric:xacml:attributes:project-tag"

    ATTRIBUTE_TYPES_AND_CATEGORIES = {
        RESOURCE_TYPE: ("http://www.w3.org/2001/XMLSchema#string",
                        "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_CPU: ("http://www.w3.org/2001/XMLSchema#integer",
                       "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_RAM: ("http://www.w3.org/2001/XMLSchema#integer",
                       "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_DISK: ("http://www.w3.org/2001/XMLSchema#integer",
                        "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_BW: ("http://www.w3.org/2001/XMLSchema#integer",
                      "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_SITE: ("http://www.w3.org/2001/XMLSchema#string",
                        "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_COMPONENT: ("http://www.w3.org/2001/XMLSchema#string",
                             "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_PEER_SITE: ("http://www.w3.org/2001/XMLSchema#string",
                             "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_MIRROR_SITE: ("http://www.w3.org/2001/XMLSchema#string",
                               "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_FACILITY_PORT: ("http://www.w3.org/2001/XMLSchema#string",
                                 "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_MEASUREMENTS: ("http://www.w3.org/2001/XMLSchema#boolean",
                                "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_PROJECT: ("http://www.w3.org/2001/XMLSchema#string",
                           "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        RESOURCE_SUBJECT: ("http://www.w3.org/2001/XMLSchema#string",
                           "urn:oasis:names:tc:xacml:3.0:attribute-category:resource"),
        ACTION_ID: ("http://www.w3.org/2001/XMLSchema#string",
                    "urn:oasis:names:tc:xacml:3.0:attribute-category:action"),
        RESOURCE_LIFETIME: ("http://www.w3.org/2001/XMLSchema#dayTimeDuration",
                            "urn:oasis:names:tc:xacml:3.0:attribute-category:action"),
        SUBJECT_ID: ("http://www.w3.org/2001/XMLSchema#string",
                     "urn:oasis:names:tc:xacml:1.0:subject-category:access-subject"),
        SUBJECT_PROJECT: ("http://www.w3.org/2001/XMLSchema#string",
                          "urn:oasis:names:tc:xacml:1.0:subject-category:access-subject"),
        PROJECT_TAG: ("http://www.w3.org/2001/XMLSchema#string",
                      "urn:oasis:names:tc:xacml:1.0:subject-category:access-subject")
    }

    METHOD_LUT = {
        ExperimentTopology: "topo",
        Node: "node",
        NetworkService: "ns",
        NetworkxASM: "asm",
        NodeSliver: "node_sliver",
        NetworkServiceSliver: "ns_sliver",
        BaseSliver: "base_sliver"
    }

    def __init__(self):
        self._attributes = defaultdict(list)
        # for now resource type is always sliver
        self._attributes[self.RESOURCE_TYPE] = ["sliver"]

    @property
    def attributes(self):
        return ViewOnlyDict(self._attributes)

    def _collect_attributes_from_node_sliver(self, sliver: NodeSliver):
        if sliver.capacities:
            self._attributes[self.RESOURCE_CPU].append(sliver.capacities.core)
            self._attributes[self.RESOURCE_RAM].append(sliver.capacities.ram)
            self._attributes[self.RESOURCE_DISK].append(sliver.capacities.disk)
        if sliver.site:
            if sliver.site not in self._attributes[self.RESOURCE_SITE]:
                self._attributes[self.RESOURCE_SITE].append(sliver.site)
        if sliver.attached_components_info:
            for c in sliver.attached_components_info.list_devices():
                self._attributes[self.RESOURCE_COMPONENT].append(str(c.get_type()))

    def _collect_attributes_from_ns_sliver(self, sliver: NetworkServiceSliver):
        if sliver.capacities:
            self._attributes[self.RESOURCE_BW].append(sliver.capacities.bw)
        if sliver.site:
            if sliver.site not in self._attributes[self.RESOURCE_SITE]:
                self._attributes[self.RESOURCE_SITE].append(sliver.site)
        if sliver.resource_type == ServiceType.PortMirror:
            if sliver.site not in self._attributes[self.RESOURCE_MIRROR_SITE]:
                self._attributes[self.RESOURCE_MIRROR_SITE].append(sliver.site)

    def _collect_attributes_from_base_sliver(self, sliver: BaseSliver):
        if isinstance(sliver, NetworkServiceSliver):
            self._collect_attributes_from_ns_sliver(sliver)
        elif isinstance(sliver, NodeSliver):
            self._collect_attributes_from_node_sliver(sliver)

    def _collect_attributes_from_topo(self, topo: ExperimentTopology):
        for n in topo.nodes.values():
            self._collect_attributes_from_node(n)
        for ns in topo.network_services.values():
            self._collect_attributes_from_ns(ns)
        for fac in topo.facilities.values():
            self._attributes[self.RESOURCE_FACILITY_PORT] = fac.name

    def _collect_attributes_from_node(self, node: Node):
        self._collect_attributes_from_node_sliver(node.get_sliver())

    def _collect_attributes_from_ns(self, ns: NetworkService):
        self._collect_attributes_from_ns_sliver(ns.get_sliver())

    def _collect_attributes_from_asm(self, asm: NetworkxASM):
        # convert to experiment topology
        asm.validate_graph()
        t = ExperimentTopology(graph_string=asm.serialize_graph())
        t.validate()
        self._collect_attributes_from_topo(t)

    def collect_resource_attributes(self, *, source: ExperimentTopology or
                                    NetworkxASM or
                                    BaseSliver or NodeSliver or
                                    NetworkServiceSliver or Node or
                                    NetworkService or None):
        """
        Take an ExperimentTopology or an ASM graph or a Sliver or a member of topology
        and extract all relevant authorization attributes, storing them as a Dict suitable for
        converting into an AuthZ request to a PDP.
        :param source: a source of attributes (ExperimentTopology, ASM or Sliver)
        :return: dictionary
        """
        assert source

        method_suffix = ResourceAuthZAttributes.METHOD_LUT.get(source.__class__)

        if not method_suffix:
            raise AuthZResourceAttributeException(f'Unsupported resource type {source.__class__} '
                                                  f'for collecting attributes')
        self.__getattribute__('_collect_attributes_from_' + method_suffix)(source)

    def set_lifetime(self, end_time: datetime):
        """
        Convert endtime of a slice to XML dayTimeDuration and set in dictionary.
        Uses utcnow to compute timedelta first.
        See http://www.datypic.com/sc/xsd/t-xsd_dayTimeDuration.html for format details
        """
        # convert to dayTimeDuration
        td = end_time - datetime.now(timezone.utc)
        days, hours, minutes, seconds = self.fromtimedelta(td)
        self._attributes[self.RESOURCE_LIFETIME].append(f'P{days}DT{hours}H{minutes}M{seconds}S')

    def set_resource_subject_and_project(self, *, subject_id: List[str] or str, project: List[str] or str):
        """
        Set resource subject (who created it) and resource project (which project it was created in
        :param subject_id: string or list of strings
        :param project: string or list of strings
        """
        if subject_id:
            if isinstance(subject_id, list):
                self._attributes[self.RESOURCE_SUBJECT].extend(subject_id)
            else:
                self._attributes[self.RESOURCE_SUBJECT].append(subject_id)

        if project:
            if isinstance(project, list):
                self._attributes[self.RESOURCE_PROJECT].extend(project)
            else:
                self._attributes[self.RESOURCE_PROJECT].append(project)

    def set_action(self, action: str):
        """
        Set the action attribute of the request
        :param action: string
        """
        if action:
            self._attributes[self.ACTION_ID].append(action)

    def set_subject_attributes(self, *, subject_id: str or None, project: str or List[str] or None,
                               project_tag: str or List[str] or None):
        """
        Set subject attributes.
        :param subject_id: string
        :param project: string or list of strings
        :param project_tag: string or list of strings
        """
        if subject_id:
            self._attributes[self.SUBJECT_ID].append(subject_id)

        if project:
            if isinstance(project, list):
                self._attributes[self.SUBJECT_PROJECT].extend(project)
            else:
                self._attributes[self.SUBJECT_PROJECT].append(project)

        if project_tag:
            if isinstance(project_tag, list):
                self._attributes[self.PROJECT_TAG].extend(project_tag)
            else:
                self._attributes[self.PROJECT_TAG].append(project_tag)

    __SECONDS_IN_DAY = 24 * 60 * 60
    __SECONDS_IN_HOUR = 60 * 60
    __SECONDS_IN_MINUTE = 60

    @staticmethod
    def fromtimedelta(td: timedelta):
        """
        Convert time delta to a tuple of days, hours, minutes and seconds
        """
        assert td.total_seconds() > 0

        ts = int(td.total_seconds())
        days = td.days
        seconds = ts - days * ResourceAuthZAttributes.__SECONDS_IN_DAY
        hours = seconds // ResourceAuthZAttributes.__SECONDS_IN_HOUR
        seconds %= ResourceAuthZAttributes.__SECONDS_IN_HOUR
        minutes = seconds // ResourceAuthZAttributes.__SECONDS_IN_MINUTE
        seconds %= ResourceAuthZAttributes.__SECONDS_IN_MINUTE
        return days, hours, minutes, seconds

    def transform_to_pdp_request(self, *, as_json: bool = True,
                                 return_policy_id_list: bool = False,
                                 combined_decision: bool = False) -> Dict[str, Any] or str:
        """
        Transform the collected attributes into a proper PDP request.
        Depending on as_json parameter returns a JSON string or a complex dict
        :param as_json: bool (default True)
        :param return_policy_id_list: bool (default False)
        :param combined_decision: bool (default False)
        """
        ret = {"Request": {
                "ReturnPolicyIdList": return_policy_id_list,
                "CombinedDecision": combined_decision,
                "Category": [
                    { "CategoryId": "urn:oasis:names:tc:xacml:3.0:attribute-category:resource",
                      "Attribute": []},
                    { "CategoryId": "urn:oasis:names:tc:xacml:3.0:attribute-category:action",
                      "Attribute": []},
                    { "CategoryId": "urn:oasis:names:tc:xacml:1.0:subject-category:access-subject",
                      "Attribute": []}
                ]}}

        cat_list = ret["Request"]["Category"]
        for k, v in self._attributes.items():
            attribute_dict = {"IncludeInResult": False, "Value": v, "AttributeId": k,
                              "DataType": self.ATTRIBUTE_TYPES_AND_CATEGORIES[k][0]}
            cat_id = self.ATTRIBUTE_TYPES_AND_CATEGORIES[k][1]
            for atcat in cat_list:
                if atcat["CategoryId"] == cat_id:
                    atcat["Attribute"].append(attribute_dict)

        if as_json:
            return json.dumps(ret)
        else:
            return ret

    def __str__(self):
        return str(self.attributes)


class AuthZResourceAttributeException(Exception):

    def __init__(self, msg: str):
        super().__init__(f"AuthZResourceAttributeException: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Abstract definition of ASM (Abstract Slice Model) functionality
"""

from typing import List, Tuple
from abc import ABCMeta, abstractmethod

import json

from fim.graph.abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException


class ABCASMPropertyGraph(ABCPropertyGraph, metaclass=ABCMeta):
    """
    Interface for ASM
    """

    @abstractmethod
    def __init__(self, *, graph_id=str, importer, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_all_network_nodes') and
                callable(subclass.get_all_network_nodes) or NotImplemented)

    @abstractmethod
    def check_node_name(self, *, node_id: str, label: str, name: str) -> bool:
        """
        Check if a node with this ID of this class/label has this name
        :param node_id:
        :param label:
        :param name:
        :return:
        """

    @abstractmethod
    def find_node_by_name(self, *, node_name: str, label: str) -> str:
        """
        Get node id of node based on its name and class/label. Throw
        exception if multiple matches found.
        :param node_name:
        :param label: node label or class of the node
        :return:
        """

    def set_mapping(self, *, node_id: str, to_graph_id: str, to_node_id: str) -> None:
        """
        Create a mapping from a node to another graph, node
        :param node_id:
        :param to_graph_id:
        :param to_node_id:
        :return:
        """
        assert node_id is not None
        assert to_graph_id is not None
        assert to_node_id is not None

        # save tuple as JSON list of guids
        l = [to_graph_id, to_node_id]
        jsonl = json.dumps(l)
        self.update_node_property(node_id=node_id, prop_name=ABCASMPropertyGraph.PROP_NODE_MAP,
                                  prop_val=jsonl)

    def get_mapping(self, *, node_id: str) -> Tuple[str, str] or None:
        """
        Retrieve a mapping, if exists for this node to another graph, node
        :param node_id:
        :return: graph_id, node_id tuple or None
        """
        assert node_id is not None
        # retrieve tuple as a json 2-member list
        _, props = self.get_node_properties(node_id=node_id)
        jsonl = props.get(ABCASMPropertyGraph.PROP_NODE_MAP, None)
        if jsonl is not None:
            l = json.loads(jsonl)
            assert(len(l) == 2)
            return tuple(l)
        return None

    def find_node_by_name_as_child(self, *, node_name: str, label: str, rel: str, parent_node_id: str) -> str or None:
        """
        Get node id of node based on its name, class label, as child of a parent node via a specified relationship
        or None
        :param node_name: 
        :param label: 
        :param rel: 
        :param parent_node_id: 
        :return: 
        """
        assert node_name is not None
        assert label is not None
        assert rel is not None
        assert parent_node_id is not None

        neighbs = self.get_first_neighbor(node_id=parent_node_id, rel=rel, node_label=label)
        for n in neighbs:
            _, props = self.get_node_properties(node_id=n)
            if props.get(ABCPropertyGraph.PROP_NAME, None) is not None and \
                props[ABCPropertyGraph.PROP_NAME] == node_name:
                return n
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find node with name {node_name} "
                                              f"class {label} as child of {parent_node_id}")

    def get_all_network_node_components(self, parent_node_id: str) -> List[str]:
        """
        Return a list of components, children of a prent (presumably network node)
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        # check that parent is a NetworkNode
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_Component)

    def get_all_network_node_or_component_nss(self, parent_node_id: str) -> List[str]:
        assert parent_node_id is not None
        # check that parent is a NetworkNode or Component
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels and \
            ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode or Component")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_NetworkService)

    def find_component_by_name(self, *, parent_node_id: str, component_name: str) -> str:

        assert component_name is not None
        component_id_list = self.get_all_network_node_components(parent_node_id=parent_node_id)
        for cid in component_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == component_name:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find component with name {component_name}")

    def find_ns_by_name(self, *, parent_node_id: str, nsname: str) -> str:

        assert nsname is not None

        ns_id_list = self.get_all_network_node_or_component_nss(parent_node_id=parent_node_id)
        for cid in ns_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == nsname:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find NetworkService with name {nsname}")

    def find_connection_point_by_name(self, *, parent_node_id: str, iname: str) -> str:

        assert iname is not None

        if_id_list = self.get_all_ns_or_link_connection_points(link_id=parent_node_id)
        for cid in if_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == iname:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find ConnectionPoint with name {iname}")




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2021 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Ilya Baldin (ibaldin@renci.org)
from fim.slivers.capacities_labels import Labels


class Gateway:

    def __init__(self, lab: Labels or None):
        """
        Labels specify IPv4 or IPv6 subnet, IPv4 or IPv6 gateway address and an optional MAC.
        Specify the following fields on Labels object:
        - ipv4_subnet and ipv4
        or
        - ipv6_subnet and ipv6
        and an optional mac
        :param lab:
        """
        if lab is None:
            self.lab = None
            return
        # labels must only specify IPv4 or v6 subnet and optional mac
        if lab.ipv4_subnet is not None and lab.ipv4 is not None:
            self.lab = Labels(ipv4_subnet=lab.ipv4_subnet, ipv4=lab.ipv4)
        elif lab.ipv6_subnet is not None and lab.ipv6 is not None:
            self.lab = Labels(ipv6_subnet=lab.ipv6_subnet, ipv6=lab.ipv6)
        else:
            raise GatewayException('Gateway must specify IPv4 or IPv6 subnet, gateway address and an optional MAC')
        if lab.mac is not None:
            self.lab.mac = lab.mac

    @property
    def gateway(self) -> str:
        if self.lab is not None:
            return self.lab.ipv4 if self.lab.ipv4 is not None else self.lab.ipv6
        return None

    @property
    def subnet(self) -> str:
        if self.lab is not None:
            return self.lab.ipv4_subnet if self.lab.ipv4_subnet is not None else self.lab.ipv6_subnet
        return None

    @property
    def mac(self) -> str:
        if self.lab is not None:
            return self.lab.mac
        return None

    def to_json(self) -> str:
        if self.lab is not None:
            return self.lab.to_json()
        return None

    @classmethod
    def from_json(cls, json_string: str):
        return Gateway(Labels.from_json(json_string))

    def __str__(self):
        ar = list()
        if self.lab is None:
            return ""
        if self.lab.ipv4_subnet is not None:
            ar.append("IPv4 subnet: " + self.lab.ipv4_subnet + " GW: " + self.lab.ipv4)
        else:
            ar.append("IPv6: " + self.lab.ipv6_subnet + " GW: " + self.lab.ipv6)

        if self.lab.mac is not None:
            ar.append("(MAC: " + self.lab.mac + ")")
        return " ".join(ar)

    def __repr__(self):
        return self.__str__()


class GatewayException(Exception):

    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Gateway exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from abc import ABC
from typing import Any

import json


class MeasurementData:
    """
    Measurement data attaches as a property on any graph node and is represented as an opaque JSON blob
    """
    MF_SIZE = 10240

    def __init__(self, data: Any or None):
        """
        data has to be a valid JSON string
        :param data:
        """

        if data is not None and isinstance(data, str):
            if len(data) > self.MF_SIZE:
                raise MeasurementDataError(f'MeasurementData JSON string is too long ({len(data)} > {self.MF_SIZE}B')
            try:
                json.loads(data)
            except json.JSONDecodeError:
                raise MeasurementDataError(f'Unable to decode measurement data {data} as valid JSON ')
            self._data = data
        elif data is not None:
            try:
                self._data = json.dumps(data)
                if len(self._data) > 1024*1024:
                    raise MeasurementDataError(f'MeasurementData object is too large: '
                                               f'{len(self._data)} > {self.MF_SIZE}B')
            except TypeError:
                raise MeasurementDataError(f'Unable to encode data as valid JSON')
        else:
            self._data = "{}"

    @property
    def data(self):
        if self._data is not None:
            return json.loads(self._data)
        return self._data

    def __str__(self):
        return str(self._data)

    def __repr(self):
        return str(self)


class MeasurementDataError(Exception):

    def __init__(self, msg):
        super().__init__(f'MeasurementDataError: invalid measurement data due to {msg}')


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List, Dict, Tuple
import json
import re

from abc import ABC, abstractmethod

import requests
import urllib.parse

from fim.graph.abc_property_graph_constants import ABCPropertyGraphConstants


class JSONField(ABC):

    @classmethod
    def update(cls, lab, **kwargs):
        """
        quasi-copy constructor if kwargs are ommitted,
        otherwise also sets additional fields. Creates a new
        instance of appropriate subclass and returns it.
        DOES NOT UPDATE IN PLACE!
        :param lab:
        :param kwargs:
        :return:
        """
        assert isinstance(lab, JSONField)
        inst = lab.__class__()
        for k, v in lab.__dict__.items():
            inst.__setattr__(k, v)
        inst._set_fields(**kwargs)
        return inst

    @abstractmethod
    def _set_fields(self, **kwargs):
        """
        Abstract private set_fields method
        :param kwargs:
        :return:
        """

    def to_json(self) -> str:
        """
        Dumps to JSON the __dict__ of the instance. Be careful as the fields in this
        class should only be those that can be present in JSON output.
        If there are no values in the object, returns empty string.
        :return:
        """
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        if len(d) == 0:
            return ''
        return json.dumps(d, skipkeys=True, sort_keys=True)

    @classmethod
    def from_json(cls, json_string: str):
        """
        Set fields from json string and returns a new object
        :param json_string:
        :return: object
        """
        if json_string is None or len(json_string) == 0 or json_string == ABCPropertyGraphConstants.NEO4j_NONE:
            return None
        d = json.loads(json_string)
        ret = cls()
        ret._set_fields(**d)
        return ret

    def to_dict(self) -> Dict[str, str] or None:
        """
        Convert to a dictionary skipping empty fields. Returns None
        if all fields are empty
        :return:
        """
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        if len(d) == 0:
            return None
        return d

    def __repr__(self):
        return self.to_json()

    def __str__(self):
        return self.to_json()

    def list_fields(self):
        l = list(self.__dict__.keys())
        l.sort()
        return l


class Capacities(JSONField):
    """
    Implements basic capacity field handling - encoding and decoding
    from JSON dictionaries of properties. Only ints are allowed.
    """
    UNITS = {'cpu': '', 'unit': '',
             'core': '', 'ram': 'G',
             'disk': 'G', 'bw': 'Gbps',
             'burst_size': 'Mbits',
             'mtu': 'B'}

    def __init__(self, **kwargs):
        self.cpu = 0
        self.core = 0
        self.ram = 0
        self.disk = 0
        self.bw = 0
        self.burst_size = 0
        self.unit = 0
        self.mtu = 0
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        """
        Universal integer setter for all fields.
        Values should be non-negative integers. Throws a CapacityException
        if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            if v is not None:
                assert v >= 0
                assert isinstance(v, int)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise CapacityException(f"Unable to set field {k} of capacity, no such field available "
                                        f"{[k for k in self.__dict__.keys()]}")
        return self

    def __add__(self, other):
        assert isinstance(other, Capacities)
        ret = Capacities()
        for f, v in self.__dict__.items():
            ret.__dict__[f] = self.__dict__[f] + other.__dict__[f]

        return ret

    def __sub__(self, other):
        assert isinstance(other, Capacities)
        ret = Capacities()

        for f, v in self.__dict__.items():
            ret.__dict__[f] = self.__dict__[f] - other.__dict__[f]

        return ret

    def __gt__(self, other):
        assert isinstance(other, Capacities)
        for f, v in self.__dict__.items():
            if v < other.__dict__[f]:
                return False
        return True

    def __lt__(self, other):
        assert isinstance(other, Capacities)
        for f, v in self.__dict__.items():
            if v > other.__dict__[f]:
                return False
        return True

    def __eq__(self, other):
        assert isinstance(other, Capacities)
        for f, v in self.__dict__.items():
            if v != other.__dict__[f]:
                return False
        return True

    def negative_fields(self) -> List[str]:
        """
        returns list of fields that are negative
        :return:
        """
        ret = list()
        for f, v in self.__dict__.items():
            if v < 0:
                ret.append(f)

        return ret

    def positive_fields(self, fields: List[str] or str) -> bool:
        """
        Return true if indicated fields are positive >0
        :param fields: string or list of strings
        :return:
        """
        assert fields is not None
        if isinstance(fields, str):
            fields = [fields]
        for f in fields:
            if self.__dict__[f] <= 0:
                return False
        return True

    def __str__(self):
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        if len(d) == 0:
            return ''
        ret = "{ "
        for i, v in d.items():
            ret = ret + i + ": " + f'{v:,} ' + self.UNITS[i] + ", "
        return ret[:-2] + "}"


class FreeCapacity:
    """
    This class takes two capacities objects (what is the total
    available and what is allocated) and allows accessing what is
    free (the difference between the two).
    """
    def __init__(self, *, total: Capacities, allocated: Capacities):
        assert total is not None
        if allocated is None:
            allocated = Capacities()
        self.total = total
        self.free = total - allocated

    def __str__(self):
        d2 = self.total.__dict__.copy()
        d1 = self.free.__dict__.copy()

        for k in self.free.__dict__:
            if d1[k] == 0 and d2[k] == 0:
                d1.pop(k)
                d2.pop(k)
        if len(d1) == 0:
            return ''
        ret = '{ '
        for k in d1:
            ret = ret + k + ": " + f'{d1[k]:,}' + "/" + f'{d2[k]:,} ' + Capacities.UNITS[k] + ", "
        return ret[:-2] + '}'

    def __getattr__(self, item):
        """
        Provide access to same fields as capacities, computed as a difference between total and allocated
        i.e. available or free
        """
        return self.free.__getattribute__(item)


class CapacityHints(JSONField):
    """
    Capacity hints are strings representing something about the capacity of a sliver,
    e.g. an instance size name. They have to be mappable to actual capacities
    via e.g. a catalog.
    """
    def __init__(self, **kwargs):
        self.instance_type = None
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values should be strings.
        Throws a LabelException if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings
            assert isinstance(v, str)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise CapacityHintException(f"Unable to set field {k} of capacity hints, no such field available")
        # to support call chaining
        return self

    def __add__(self, other):
        assert isinstance(other, Labels)
        raise RuntimeError("Not Implemented")

    def __str__(self):
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None:
                d.pop(k)
        if len(d) == 0:
            return ''
        ret = "{ "
        for i, v in d.items():
            ret = ret + i + ": " + str(v) + ", "
        return ret[:-2] + "}"


class Labels(JSONField):
    """
    Class implementing various encodings of labels field, encoding
    and decoding from JSON dictionaries of properties
    """
    VALIDATORS = {
        'bdf': ('[0-9a-fA-F]{1,4}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}.[0-9a-fA-F]+', "0000:00:00.0"),
        'mac': ('([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}', "00:11:22:33:44:55"),
        'ipv4': (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)',
                 "192.168.1.1"),
        'ipv4_range': (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)-'
                      r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)',
                       "192.168.1.1-192.168.1.10"),
        'ipv4_subnet': (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)/[\d]{1,2}',
                        "192.168.1.0/24"),
        'ipv6': (r'(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}',
                 "2001:0db8:85a3:0000:0000:8a2e:0370:7334"),
        'ipv6_range': (r'(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}-(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}',
                      "2001:0db8:85a3:0000:0000:8a2e:0370:7334-2001:0db8:85a3:0000:0000:8a2e:0370:8334"),
        # we allow fewer than 128 (and not necessarily 64) bits to be specified, unlike ipv6 address
        # where we require the full 128 even if some of them are just ':::'
        'ipv6_subnet': (r'(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}/[\d]{1,2}',
                        "2001:0db8:85a3:0000:0000/48"),
        'asn': (r'[\d]+', "12345"),
        'vlan': (r'[\d]{1,4}', "1234"),
        'vlan_range': (r'[\d]{1,4}-[\d]{1,4}', "100-200"),
        'inner_vlan': (r'[\d]{1,4}', "1234")
    }
    LAMBDA_VALIDATORS = {
        'vlan': ((lambda v: True if 0 < int(v) <= 4096 else False), "1-4096"),
        'inner_vlan': ((lambda v: True if 0 < int(v) <= 4096 else False), "1-4096"),
        'vlan_range': ((lambda v: True if 0 < int(v.split('-')[0]) <= 4096 and
                                         0 < int(v.split('-')[1]) <= 4096 and
                                         int(v.split('-')[0]) < int(v.split('-')[1]) else False),
                       "1-4096"),
        'asn': ((lambda a: True if 0 < int(a) < 65536 else False), "1-65535")
    }

    def __init__(self, **kwargs):
        self.bdf = None
        self.mac = None
        self.ipv4 = None
        self.ipv4_range = None
        self.ipv4_subnet = None
        self.ipv6 = None
        self.ipv6_range = None
        self.ipv6_subnet = None
        self.asn = None
        self.vlan = None
        self.vlan_range = None
        self.inner_vlan = None
        self.instance = None
        self.instance_parent = None
        self.local_name = None
        self.local_type = None
        self.device_name = None
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values should be strings or lists of strings.
        Throws a LabelException if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                if self.VALIDATORS.get(k, None) is not None:
                    if isinstance(v, list):
                        for i in v:
                            matches = re.match('^' + self.VALIDATORS[k][0] + '$', i)
                            if matches is None:
                                raise LabelException(f'Provided label value {i} for {k} does not match the allowed '
                                                     f'regular expression {self.VALIDATORS[k][0]}, valid example is '
                                                     f'{self.VALIDATORS[k][1]}')
                    else:
                        matches = re.match('^' + self.VALIDATORS[k][0] + '$', v)
                        if matches is None:
                            raise LabelException(f'Provided label value {v} for {k} does not match the allowed '
                                                 f'regular expression {self.VALIDATORS[k][0]}, valid example is '
                                                 f'{self.VALIDATORS[k][1]}')
                if self.LAMBDA_VALIDATORS.get(k, None) is not None:
                    if isinstance(v, list):
                        for i in v:
                            res = self.LAMBDA_VALIDATORS[k][0](i)
                            if res is False:
                                raise LabelException(f'Provided label value {i} for {k} must be in a valid '
                                                     f'range {self.LAMBDA_VALIDATORS[k][1]}')
                    else:
                        res = self.LAMBDA_VALIDATORS[k][0](v)
                        if res is False:
                            raise LabelException(f'Provided label value {v} for {k} must be in a valid '
                                                 f'range {self.LAMBDA_VALIDATORS[k][1]}')

                self.__setattr__(k, v)
            except AttributeError:
                raise LabelException(f"Unable to set field {k} of labels, no such field available "
                                     f"{[k for k in self.__dict__.keys()]}")
        # to support call chaining
        return self

    def __add__(self, other):
        assert isinstance(other, Labels)
        # FIXME: does ADD mean finding a union of label sets per type?
        raise RuntimeError("Not Implemented")

    def __str__(self):
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None:
                d.pop(k)
        if len(d) == 0:
            return ''
        ret = "{ "
        for i, v in d.items():
            ret = ret + i + ": " + str(v) + ", "
        return ret[:-2] + "}"


class ReservationInfo(JSONField):
    """
    Reservation info structure for ASM sliver objects
    """

    def __init__(self, **kwargs):
        self.reservation_id = None
        self.reservation_state = None
        self.error_message = None
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values should be strings or lists of strings.
        Throws a ReservationInfoException if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise ReservationInfoException(f"Unable to set field {k} of reservation info, no such field "
                                               f"available {[k for k in self.__dict__.keys()]}")
        # to support call chaining
        return self


class StructuralInfo(JSONField):
    """
    Structural info on the for sliver objects (things like - is it a stitching node,
    what is parent graph or subgraph)
    """
    def __init__(self, **kwargs):
        self.sub_graph_id = None
        self.parent_graph_id = None
        self.adm_graph_ids = None
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values are strings or boolean.
        Throws a
        :param kwargs:
        :return:
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise StructuralInfoException(f"Unable to set field {k} of structural info, no such field available")
        # to support call chaining
        return self


class Location(JSONField):
    """
    Location representation (as address, coordinates or any other useful way)
    """
    def __init__(self, **kwargs):
        self.postal = None
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        """
        Universal setter for location fields. Values are strings.
        :param kwargs:
        :return:
        """
        for k, v in kwargs.items():
            assert v is not None
            assert isinstance(v, str)
            try:
                # will throw exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise LocationException(f"Unable to set field {k} of location, no such field available")
        return self

    def to_latlon(self) -> Tuple[float, float]:
        """
        Return a tuple of floats indicating Lat/Lon of the location. Uses Nomatim OpenStreetMaps
        service.
        :return:
        """
        url = 'https://nominatim.openstreetmap.org/search/' + urllib.parse.quote(self.postal) + '?format=json'
        # per terms of service set user agent
        headers = {'User-Agent': 'FABRIC FIM Utility'}
        response = requests.get(url, headers)
        if response.status_code != 200:
            raise LocationException(f"Unable to convert address to Lat/Lon via OpenStreetmaps due "
                                    f"to: {response.reason}")
        try:
            response_json = response.json()
            if not isinstance(response_json, list):
                raise ValueError
            if len(response_json) < 1:
                raise ValueError
        except ValueError:
            raise LocationException(f"Unable to interpret response from OpenStreetmaps for address {self.postal}")

        return float(response_json[0]['lat']), float(response_json[0]['lon'])


class Flags(JSONField):
    """
    JSON-ified representation of various flags that can be attached to slivers
    """
    def __init__(self, **kwargs):
        self.auto_config = False
        self._set_fields(**kwargs)

    def _set_fields(self, **kwargs):
        for k, v in kwargs.items():
            assert v is not None
            assert isinstance(v, bool)
            try:
                # will throw exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise FlagException(f"Unable to set field {k} of location, no such field available")
        return self

    def to_json(self) -> str:
        """
        Dumps to JSON the __dict__ of the instance. Be careful as the fields in this
        class should only be those that can be present in JSON output.
        Specialized for boolean values from the generic implementation in JSONFields
        :return:
        """
        d = self.__dict__.copy()
        return json.dumps(d, skipkeys=True, sort_keys=True)


class CapacityException(Exception):
    """
    Exception with a capacity
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Delegation exception: {msg}")


class CapacityHintException(Exception):
    """
    Exception with a capacity
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Delegation exception: {msg}")


class LabelException(Exception):
    """
    Exception with a label
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Label exception: {msg}")


class ReservationInfoException(Exception):
    """
    Exception with a reservation info
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"ReservationInfo exception: {msg}")


class StructuralInfoException(Exception):
    """
    Exception with a structural info
    """

    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"ReservationInfo exception: {msg}")


class LocationException(Exception):
    """
    Exception with location
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Location exception: {msg}")


class FlagException(Exception):
    """
    Exception with flags
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Flag exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
import logging
from typing import Tuple, Any, List, Set

import uuid

from fim.view_only_dict import ViewOnlyDict

from .model_element import ModelElement, ElementType, TopologyException

from fim.graph.abc_property_graph import ABCPropertyGraph
from fim.user.interface import Interface, InterfaceType
from fim.user.link import Link, LinkType
from fim.slivers.network_service import NetworkServiceSliver, ServiceType, NSLayer, ERO, PathInfo, MirrorDirection
from fim.slivers.gateway import Gateway


class NetworkService(ModelElement):
    """
    A network service object in a topology that can pass packets between interfaces.
    Unlike a link, network service owns its interfaces. Typical topology fragment is
    component - network service - interface - link - interface - network service
    In addition to public methods the following calls
    return various dictionaries or lists:
    service.interface_list - a list of interfaces attached to this service
    """
    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING,
                 parent_node_id: str = None,
                 interfaces: List[Interface] = None,
                 nstype: ServiceType = None, technology: str = None,
                 **kwargs):
        """
        Don't call this method yourself, call topology.add_network_service()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to exist or new should be created
        :param parent_node_id: node_id of the parent Node if any (for new components)
        :param interfaces: list of interface objects to connect
        :param nstype: service type if new
        :param technology: service technology
        :param kwargs: any additional properties
        """
        assert name is not None
        assert topo is not None

        site = None
        if etype == ElementType.NEW:
            # node id myst be specified for new nodes in substrate topologies
            if node_id is None:
                node_id = str(uuid.uuid4())
            super().__init__(name=name, node_id=node_id, topo=topo)
            if nstype is None:
                raise TopologyException("When creating new services you must specify ServiceType")
            # cant use isinstance as it would create circular import dependencies
            # We do more careful checks for ExperimentTopologies, but for substrate we let things loose
            if str(self.topo.__class__) == "<class 'fim.user.topology.ExperimentTopology'>" and interfaces:
                sites = self.__validate_nstype_constraints(nstype, interfaces)

                # if a single-site service
                if len(sites) == 1:
                    site = sites.pop()

            sliver = NetworkServiceSliver()
            sliver.node_id = self.node_id
            sliver.set_name(self.name)
            sliver.set_type(nstype)
            sliver.set_site(site)
            # set based on service type
            sliver.set_layer(NetworkServiceSliver.ServiceConstraints[nstype].layer)
            sliver.set_technology(technology)
            sliver.set_properties(**kwargs)
            self.topo.graph_model.add_network_service_sliver(parent_node_id=parent_node_id, network_service=sliver)
            self._interfaces = list()
            if interfaces is not None and len(interfaces) > 0:
                for i in interfaces:
                    # run through guardrails, then connect
                    self.__service_guardrails(sliver, i)
                    self.__connect_interface(interface=i)
        else:
            assert node_id is not None
            super().__init__(name=name, node_id=node_id, topo=topo)
            # check that this node exists
            existing_node_id = self.topo.\
                graph_model.find_node_by_name(node_name=name,
                                              label=ABCPropertyGraph.CLASS_NetworkService)
            if existing_node_id != node_id:
                raise TopologyException(f'Service name {name} node id does not match the expected node id.')
            # collect a list of interface nodes it attaches to
            interface_list = self.topo.graph_model.get_all_ns_or_link_connection_points(link_id=self.node_id)
            name_id_tuples = list()
            # need to look up their names - a bit inefficient, need to think about this /ib
            for iff in interface_list:
                _, props = self.topo.graph_model.get_node_properties(node_id=iff)
                name_id_tuples.append((props[ABCPropertyGraph.PROP_NAME], iff))
            self._interfaces = [Interface(node_id=tup[1], topo=topo, name=tup[0]) for tup in name_id_tuples]

    @property
    def type(self):
        return self.get_property('type') if self.__dict__.get('topo', None) is not None else None

    @property
    def site(self):
        return self.get_property('site') if self.__dict__.get('topo', None) is not None else None

    @site.setter
    def site(self, value: str):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('site', value)

    @property
    def technology(self):
        return self.get_property('technology') if self.__dict__.get('topo', None) is not None else None

    @property
    def layer(self):
        return self.get_property('layer') if self.__dict__.get('topo', None) is not None else None

    @property
    def controller_url(self):
        return self.get_property('controller_url') if self.__dict__.get('topo', None) is not None else None

    @controller_url.setter
    def controller_url(self, value: str):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('controller_url', value)

    @property
    def ero(self):
        return self.get_property('ero') if self.__dict__.get('topo', None) is not None else None

    @ero.setter
    def ero(self, value: ERO):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('ero', value)

    @property
    def path_info(self):
        return self.get_property('path_info') if self.__dict__.get('topo', None) is not None else None

    @path_info.setter
    def path_info(self, value: PathInfo):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('path_info', value)

    @property
    def gateway(self):
        return self.get_property('gateway') if self.__dict__.get('topo', None) is not None else None

    @gateway.setter
    def gateway(self, gateway: Gateway):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('gateway', gateway)

    def get_sliver(self) -> NetworkServiceSliver:
        """
        Get a deep sliver representation of this node from graph
        :return:
        """
        return self.topo.graph_model.build_deep_ns_sliver(node_id=self.node_id)

    def __validate_nstype_constraints(self, nstype, interfaces) -> Set[str]:
        """
        Validate service constraints as encoded for each services (number of interfaces, instances, sites).
        Note that interfaces is a list of interfaces belonging to nodes!
        :param nstype:
        :param interfaces:
        :return: a list of sites the service covers
        """

        # check the number of instances of this service
        if NetworkServiceSliver.ServiceConstraints[nstype].num_instances != NetworkServiceSliver.NO_LIMIT:
            services = self.topo.graph_model.get_all_nodes_by_class_and_type(label=ABCPropertyGraph.CLASS_NetworkService,
                                                                             ntype=str(nstype))
            if len(services) + 1 > NetworkServiceSliver.ServiceConstraints[self.type].num_instances:
                raise TopologyException(f"Service type {nstype} cannot have {len(services) + 1} instances. "
                                        f"Limit: {NetworkServiceSliver.ServiceConstraints[nstype].num_instances}")
        # check the number of interfaces
        if NetworkServiceSliver.ServiceConstraints[nstype].num_interfaces != NetworkServiceSliver.NO_LIMIT:
            if len(interfaces) > NetworkServiceSliver.ServiceConstraints[nstype].num_interfaces:
                raise TopologyException(f"Service of type {nstype} cannot have {len(interfaces)} interfaces. "
                                        f"Limit: {NetworkServiceSliver.ServiceConstraints[nstype].num_interfaces}")
        sites = set()
        # check the number of sites spanned by this service
        if NetworkServiceSliver.ServiceConstraints[nstype].num_sites != NetworkServiceSliver.NO_LIMIT:
            # trace ownership of each interface and count the sites involved
            for interface in interfaces:
                owner = self.topo.get_owner_node(interface)
                if owner is None:
                    print(f'Interface {interface=} has no owner')
                sites.add(owner.site)

            if len(sites) > NetworkServiceSliver.ServiceConstraints[nstype].num_sites:
                raise TopologyException(f"Service of type {nstype} cannot span {len(sites)} sites. "
                                        f"Limit: {NetworkServiceSliver.ServiceConstraints[nstype].num_sites}.")
        return sites

    def validate_constraints(self, interfaces):
        """
        Validate service constraints - number of sites, interfaces, instances, properties
        """

        nstype = self.type
        self.__validate_nstype_constraints(nstype, interfaces)

        # check properties
        req_props = NetworkServiceSliver.ServiceConstraints[nstype].required_properties
        forb_props = NetworkServiceSliver.ServiceConstraints[nstype].forbidden_properties
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        ns_sliver = self.topo.graph_model.network_service_sliver_from_graph_properties_dict(node_properties)
        for rp in req_props:
            if not ns_sliver.get_property(rp):
                raise TopologyException(f"Service of type {nstype} must have property {rp} set")
        for fp in forb_props:
            if ns_sliver.get_property(fp):
                raise TopologyException(f"Service of type {nstype} must NOT have property {fp} set")

        # check interface types that are required to attach
        rit = NetworkServiceSliver.ServiceConstraints[nstype].required_interface_types
        if len(rit) > 0:
            for i in interfaces:
                if not i.type in rit:
                    raise TopologyException(f"Service of type {nstype} must use one of the following interface types:"
                                            f"{rit} instead of {i.type}")

    @staticmethod
    def __service_guardrails(sliver: NetworkServiceSliver, interface: Interface):
        """
        Checks if this interface can be added to this service for various reasons related to e.g.
        service implementation constraints (that can be temporary and change from release to release).
        :param sliver:
        :param interface:
        :return:
        """
        # - L2P2P service does not work for shared ports
        # - L2S2S needs to warn that it may not work if the VMs with shared ports land on the same worker
        if sliver.get_type() == ServiceType.L2PTP and \
            interface.type == InterfaceType.SharedPort:
            raise TopologyException(f"Unable to connect interface {interface.name} to service {sliver.get_name()}: "
                                    f"L2P2P service currently doesn't support shared interfaces")

    def __connect_interface(self, *, interface: Interface):
        """
        Connect a (compute or switch) node interface to network service by transparently
        creating a peer service interface and a link between them
        :param interface
        :return:
        """
        assert interface is not None
        assert isinstance(interface, Interface)

        # we can only connect interfaces connected to (compute or switch) nodes,
        parent = self.topo.get_owner_node(interface)
        if parent is None:
            raise TopologyException(f'Interface {interface} is not owned by a node, as expected.')
        # we can only connect interfaces that aren't already connected
        peer_ids = self.topo.graph_model.find_peer_connection_points(node_id=interface.node_id)
        if peer_ids is not None:
            raise TopologyException(f'Interface {interface} is already connected to another service.')
        # create a peer interface, create a link between them
        peer_if = Interface(name='-'.join([parent.name, interface.name]),
                            parent_node_id=self.node_id,
                            etype=ElementType.NEW, topo=self.topo, itype=InterfaceType.ServicePort)
        # link type is determined by the type of interface = L2Path for shared, Patch for Dedicated
        if interface.type == InterfaceType.SharedPort:
            ltype = LinkType.L2Path
        else:
            ltype = LinkType.Patch

        peer_link = Link(name=peer_if.name + '-link', topo=self.topo, etype=ElementType.NEW,
                         interfaces=[interface, peer_if], ltype=ltype)
        self._interfaces.append(peer_if)

    def disconnect_interface(self, *, interface: Interface) -> None:
        """
        Disconnect a node interface from the network service.
        Transparently remove peer service interface and link between them.
        :param interface:
        :return:
        """
        assert interface is not None

        peers = interface.get_peers()
        if peers is None or len(peers) == 0:
            return

        if len(peers) > 1:
            raise TopologyException(f'List of ServicePort peers for interface {interface} is more than one')

        self.topo.graph_model.remove_cp_and_links(node_id=peers[0].node_id)
        # remove from interface list as well
        self._interfaces = list(filter((lambda x: x.node_id != peers[0].node_id), self._interfaces))

    def add_interface(self, *, name: str, node_id: str = None, itype: InterfaceType = InterfaceType.TrunkPort,
                      **kwargs):
        """
        Add an interface to network service
        :param name:
        :param node_id:
        :param itype: interface type e.g. TrunkPort, AccessPort or VINT
        :param kwargs: additional parameters
        :return:
        """
        assert name is not None

        # check uniqueness
        if name in self.__list_interfaces().keys():
            raise TopologyException('Interface names must be unique within a switch fabric')
        iff = Interface(name=name, node_id=node_id, parent_node_id=self.node_id,
                        etype=ElementType.NEW, topo=self.topo, itype=itype,
                        **kwargs)
        return iff

    def remove_interface(self, *, name: str) -> None:
        """
        Remove an ServicePort interface from the network service.
        :param name:
        :return:
        """
        assert name is not None
        # cant use isinstance as it would create circular import dependencies
        if str(self.topo.__class__) == "<class 'fim.user.topology.ExperimentTopology'>":
            raise TopologyException("Cannot remove interface from NetworkService in Experiment topology")
        node_id = self.topo.graph_model.find_connection_point_by_name(parent_node_id=self.node_id,
                                                                      iname=name)
        self.topo.graph_model.remove_cp_and_links(node_id=node_id)

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a service property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        ns_sliver = self.topo.graph_model.network_service_sliver_from_graph_properties_dict(node_properties)
        return ns_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a service property or unset of pval is None
        :param pname:
        :param pval:
        :return:
        """
        if pval is None:
            self.unset_property(pname)
            return
        service_sliver = NetworkServiceSliver()
        service_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.network_service_sliver_to_graph_properties_dict(service_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the service
        :param kwargs:
        :return:
        """
        service_sliver = NetworkServiceSliver()
        service_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.network_service_sliver_to_graph_properties_dict(service_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> Tuple[str]:
        return NetworkServiceSliver.list_properties()

    def __get_interface_by_id(self, node_id: str) -> Interface:
        """
        Get an interface of network service by its node_id, return Interface object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __get_interface_by_name(self, name: str) -> Interface:
        """
        Get an interface of network service by its name
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.topo.graph_model.find_connection_point_by_name(parent_node_id=self.node_id,
                                                                      iname=name)
        return Interface(name=name, node_id=node_id, topo=self.topo)

    def __list_interfaces(self) -> ViewOnlyDict:
        """
        List all interfaces of the network service as a dictionary
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_ns_or_link_connection_points(link_id=self.node_id)
        ret = dict()
        for nid in node_id_list:
            c = self.__get_interface_by_id(nid)
            ret[c.name] = c
        return ViewOnlyDict(ret)

    def __list_of_interfaces(self) -> Tuple[Interface]:
        """
        Return a list of all interfaces of network service
        :return:
        """
        return tuple(self._interfaces)

    @property
    def interface_list(self):
        return self.__list_of_interfaces()

    @property
    def interfaces(self):
        return self.__list_interfaces()

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        service_sliver = self.topo.graph_model.network_service_sliver_from_graph_properties_dict(node_properties)
        interface_names = [iff.name for iff in self._interfaces]
        return service_sliver.__repr__() + str(interface_names)

    def __str__(self):
        return self.__repr__()


class PortMirrorService(NetworkService):
    """
    PortMirroring service is special because interfaces in it aren't equal.
    There is the mirrored interface and there is a mirrored-to interface.
    Also it is not typically defined as part of ARMs, CBMs or (A)BQMs.
    """

    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING,
                 parent_node_id: str = None, direction: MirrorDirection = MirrorDirection.Both,
                 from_interface_name: str = None, to_interface: Interface = None,
                 **kwargs):
        if etype == ElementType.NEW:
            assert to_interface
            assert from_interface_name

            # only the 'to_interface' is actually connected to the service
            # the 'from_interface' is connected to something else
            # to_interface has to be a full-rate interface

            # make sure that to_interface is a DedicatedPort
            if to_interface.type != InterfaceType.DedicatedPort:
                raise TopologyException(f'Adding PortMirrorService {name} failed - only dedicated '
                                        f'ports belonging to SmartNICs can be attached.')
            super().__init__(name=name, node_id=node_id, topo=topo, etype=etype,
                             parent_node_id=parent_node_id, interfaces=[to_interface],
                             nstype=ServiceType.PortMirror, technology=None,
                             mirror_port=from_interface_name, mirror_direction=direction,
                             **kwargs)
        else:
            assert node_id is not None
            super().__init__(name=name, node_id=node_id, topo=topo)
            # check that this node exists
            existing_node_id = self.topo.\
                graph_model.find_node_by_name(node_name=name,
                                              label=ABCPropertyGraph.CLASS_NetworkService)
            if existing_node_id != node_id:
                raise TopologyException(f'Service name {name} node id does not match the expected node id.')
            # collect a list of interface nodes it attaches to
            interface_list = self.topo.graph_model.get_all_ns_or_link_connection_points(link_id=self.node_id)
            name_id_tuples = list()
            # need to look up their names - a bit inefficient, need to think about this /ib
            for iff in interface_list:
                _, props = self.topo.graph_model.get_node_properties(node_id=iff)
                name_id_tuples.append((props[ABCPropertyGraph.PROP_NAME], iff))
            self._interfaces = [Interface(node_id=tup[1], topo=topo, name=tup[0]) for tup in name_id_tuples]

    @property
    def mirror_port(self):
        return self.get_property('mirror_port') if self.__dict__.get('topo', None) is not None else None

    @property
    def mirror_direction(self):
        return self.get_property('mirror_direction') if self.__dict__.get('topo', None) is not None else None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from abc import ABC

import re
import json


class Tags:
    """
    A class representing a list of string tags that are no longer than a specified limit
    """
    TAG_PATTERN="^[\\w-]{1,255}$"
    compiled_pattern = re.compile(TAG_PATTERN)

    def __init__(self, *kwargs):
        """
        Provide tags as a list or individual parameters
        :param kwargs:
        """
        self.tags = list()
        for k in kwargs:
            if isinstance(k, list) or isinstance(k, tuple):
                for t in k:
                    Tags._check(t)
                    self.tags.append(t)
            else:
                Tags._check(k)
                self.tags.append(k)

    def to_json(self) -> str:
        """
        Convert list of tags to JSON
        :return:
        """
        return json.dumps(self.tags)

    def __str__(self) -> str:
        return str(self.tags)

    def __repr(self) -> str:
        return str(self)

    @classmethod
    def from_json(cls, json_string: str):
        """
        Create a Tags object from JSON string
        :return:
        """
        if json_string is None or len(json_string) == 0 or json_string == "None":
            return None

        d = json.loads(json_string)
        ret = cls(d)
        return ret

    @staticmethod
    def _check(tag: str):
        """
        Check that the tag conforms to our rules - it is a string of regular characters
        no longer than limit
        :param tag:
        :return:
        """
        if not isinstance(tag, str) or Tags.compiled_pattern.match(tag) is None:
            raise TagException(f'Tag {tag} does not match expected pattern {Tags.TAG_PATTERN}')

    def __iter__(self):
        return iter(list(self.tags))


class TagException(Exception):

    def __init__(self, msg):
        super().__init__(f'TagException: {msg}')



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import Tuple, Any
import enum
from recordclass import recordclass

from .base_sliver import BaseSliver
from .path_info import PathRepresentationType, ERO, PathInfo
from .gateway import Gateway
from .interface_info import InterfaceType


class NSLayer(enum.Enum):
    L0 = enum.auto()
    L1 = enum.auto()
    L2 = enum.auto()
    L3 = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str):
        for la in NSLayer:
            if la.name == s:
                return cls(la)
        return None


class MirrorDirection(enum.Enum):
    Both = enum.auto()
    RX_Only = enum.auto()
    TX_Only = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str):
        for md in MirrorDirection:
            if md.name == s:
                return cls(md)
        return None


class ServiceType(enum.Enum):
    """
    Possible Service types in FABRIC. Unlike links,
    Services can be instantiated on demand.
    """
    P4 = enum.auto()
    MPLS = enum.auto()
    OVS = enum.auto() # OVS service typically inside a NIC
    L2Path = enum.auto() # provider L2 path (overlap with link types -
    # it may be possible to select provider path as a service)
    L2STS = enum.auto() # FABRIC Site-to-Site service
    L2PTP = enum.auto() # FABRIC Port-to-Port service
    L2Multisite = enum.auto() # FABRIC multi-site service
    L2Bridge = enum.auto() # FABRIC L2 bridge within a site
    FABNetv4 = enum.auto() # FABRIC IPv4 routed network
    FABNetv6 = enum.auto() # FABRIC IPv6 routed network
    PortMirror = enum.auto() # FABRIC port mirroring service
    L3VPN = enum.auto() # FABRIC L3 VPN service
    VLAN = enum.auto() # a local VLAN (internal to site)

    def help(self) -> str:
        return NetworkServiceSliver.ServiceConstraints[self].desc

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


ServiceConstraintRecord = recordclass('ServiceConstraintRecord',
                                      ['layer', 'desc',
                                       'num_interfaces',
                                       'num_sites',
                                       'num_instances',
                                       'required_properties',
                                       'forbidden_properties',
                                       'required_interface_types'])


class NetworkServiceSliver(BaseSliver):

    # whenever there is no limit, num is set to 0
    NO_LIMIT = 0
    """
    Services can be limited by the number of interfaces/connection points they can connect
    The number of sites they may connect and the number of instances they may have in a slice.
    """
    ServiceConstraints = {
        ServiceType.P4: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=1,
                                                num_instances=NO_LIMIT, required_properties=['controller_url'],
                                                forbidden_properties=['mirror_port',
                                                                      'mirror_direction'],
                                                required_interface_types=[],
                                                desc='A P4 service.'),
        ServiceType.OVS: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=1,
                                                 num_instances=NO_LIMIT, required_properties=[],
                                                 forbidden_properties=['mirror_port',
                                                                       'mirror_direction'],
                                                 required_interface_types=[],
                                                 desc='An OVS generic service.'),
        ServiceType.VLAN: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=1,
                                                  num_instances=NO_LIMIT, required_properties=[],
                                                  forbidden_properties=['mirror_port',
                                                                        'mirror_direction',
                                                                        'controller_url'],
                                                  required_interface_types=[],
                                                  desc='An local VLAN service in a site.'),
        ServiceType.MPLS: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=1,
                                                  num_instances=NO_LIMIT, desc='An MPLS generic service',
                                                  required_properties=[],
                                                  forbidden_properties=['mirror_port',
                                                                        'mirror_direction',
                                                                        'controller_url'],
                                                  required_interface_types=[]),
        ServiceType.L2Path: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=2, num_sites=2,
                                                    num_instances=NO_LIMIT,
                                                    desc='A provider L2 Path e.g. from ESnet or Internet2.',
                                                    required_properties=[],
                                                    forbidden_properties=['mirror_port',
                                                                          'mirror_direction',
                                                                          'controller_url'],
                                                    required_interface_types=[]),
        ServiceType.L2STS: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=2,
                                                   num_instances=NO_LIMIT,
                                                   desc='A Site-to-Site service in FABRIC.',
                                                   required_properties=[],
                                                   forbidden_properties=['ero',
                                                                         'mirror_port',
                                                                         'mirror_direction',
                                                                         'controller_url'],
                                                   required_interface_types=[]),
        ServiceType.L2PTP: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=2, num_sites=2,
                                                   num_instances=NO_LIMIT,
                                                   desc='A Port-to-Port service in FABRIC.',
                                                   required_properties=[],
                                                   forbidden_properties=['mirror_port',
                                                                         'mirror_direction',
                                                                         'controller_url'],
                                                   required_interface_types=[InterfaceType.DedicatedPort]),
        ServiceType.L2Multisite: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=NO_LIMIT,
                                                         num_instances=NO_LIMIT,
                                                         desc='A Multi-Site L2 service in FABRIC.',
                                                         required_properties=[],
                                                         forbidden_properties=['mirror_port',
                                                                               'mirror_direction',
                                                                               'controller_url'],
                                                         required_interface_types=[]),
        ServiceType.L2Bridge: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=NO_LIMIT, num_sites=1,
                                                      num_instances=NO_LIMIT,
                                                      desc='An L2 bridge service within a single FABRIC site.',
                                                      required_properties=[],
                                                      forbidden_properties=['mirror_port',
                                                                            'mirror_direction',
                                                                            'controller_url'],
                                                      required_interface_types=[]),
        ServiceType.FABNetv4: ServiceConstraintRecord(layer=NSLayer.L3, num_interfaces=NO_LIMIT, num_sites=1,
                                                      num_instances=NO_LIMIT,
                                                      desc='A routed IPv4 (RFC1918 addressed) FABRIC network.',
                                                      required_properties=[],
                                                      forbidden_properties=['mirror_port',
                                                                            'mirror_direction',
                                                                            'controller_url'],
                                                      required_interface_types=[]),
        ServiceType.FABNetv6: ServiceConstraintRecord(layer=NSLayer.L3, num_interfaces=NO_LIMIT, num_sites=1,
                                                      num_instances=NO_LIMIT,
                                                      desc='A routed IPv6 (publicly addressed) FABRIC network.',
                                                      required_properties=[],
                                                      forbidden_properties=['mirror_port',
                                                                            'mirror_direction',
                                                                            'controller_url'],
                                                      required_interface_types=[]),
        ServiceType.PortMirror: ServiceConstraintRecord(layer=NSLayer.L2, num_interfaces=1, num_sites=1,
                                                        num_instances=NO_LIMIT,
                                                        desc='A port mirroring service in a FABRIC site.',
                                                        required_properties=['mirror_port',
                                                                             'mirror_direction', 'site'],
                                                        forbidden_properties=['controller_url'],
                                                        required_interface_types=[InterfaceType.DedicatedPort]),
        ServiceType.L3VPN: ServiceConstraintRecord(layer=NSLayer.L3, num_interfaces=NO_LIMIT, num_sites=NO_LIMIT,
                                                   num_instances=NO_LIMIT,
                                                   desc='A L3 VPN service connecting to FABRIC.',
                                                   required_properties=[],
                                                   forbidden_properties=['mirror_port',
                                                                         'mirror_direction',
                                                                         'controller_url'],
                                                   required_interface_types=[])
    }

    def __init__(self):
        super().__init__()
        self.layer = None
        self.technology = None
        self.interface_info = None
        self.allocation_constraints = None
        self.ero = None
        self.path_info = None
        self.controller_url = None
        self.site = None
        self.gateway = None
        self.mirror_port = None
        self.mirror_direction = None

    #
    # Setters are only needed for things we want users to be able to set
    #
    def get_layer(self) -> NSLayer:
        return self.layer

    def set_layer(self, layer: NSLayer):
        self.layer = layer

    def get_technology(self) -> str:
        return self.technology

    def set_technology(self, technology: str):
        self.technology = technology

    def set_allocation_constraints(self, allocation_constraints: str):
        self.allocation_constraints = allocation_constraints

    def get_allocation_constraints(self) -> str:
        return self.allocation_constraints

    def set_ero(self, ero: ERO) -> None:
        self.ero = ero

    def get_ero(self) -> ERO:
        return self.ero

    def set_path_info(self, path_info: PathInfo) -> None:
        self.path_info = path_info

    def get_path_info(self) -> PathInfo:
        return self.path_info

    def set_controller_url(self, curl: str) -> None:
        self.controller_url = curl

    def get_controller_url(self) -> None:
        return self.controller_url

    def set_site(self, site: str):
        self.site = site

    def get_site(self) -> str:
        return self.site

    def set_gateway(self, gw: Gateway):
        self.gateway = gw

    def get_gateway(self) -> Gateway:
        return self.gateway

    def set_mirror_port(self, mirror_port: str):
        self.mirror_port = mirror_port

    def get_mirror_port(self) -> str:
        return self.mirror_port

    def set_mirror_direction(self, mirror_direction: MirrorDirection):
        self.mirror_direction = mirror_direction

    def get_mirror_direction(self) -> MirrorDirection:
        return self.mirror_direction

    @staticmethod
    def type_from_str(ltype: str) -> ServiceType or None:
        if ltype is None:
            return None
        for t in ServiceType:
            if ltype == str(t):
                return t

    @staticmethod
    def layer_from_str(layer: str) -> NSLayer or None:
        if layer is None:
            return None
        for t in NSLayer:
            if layer == str(t):
                return t


class NetworkServiceInfo:

    def __init__(self):
        self.network_services = {}

    def add_network_service(self, ns_sliver: NetworkServiceSliver):
        self.network_services[ns_sliver.resource_name] = ns_sliver

    def remove_network_service(self, name: str):
        assert name is not None

        if name in self.network_services.keys():
            self.network_services.pop(name)

    def get_network_service(self, name: str):
        assert name is not None

        return self.network_services.get(name, None)

    def get_network_service_names(self):
        return list(self.network_services.keys())

    def list_services(self):
        return list(self.network_services.values())

